import inspect
import unicodedata

from django.utils.functional import cached_property
from django.utils.html import escape
from django.utils.safestring import SafeData, mark_safe
from django.utils.text import normalize_newlines


def remove_accent(string_with_diacritics):
    """
    Removes a diacritics from a given string"
    """
    return unicodedata.normalize('NFKD', string_with_diacritics).encode('ASCII', 'ignore').decode('ASCII')


def get_class_method(cls_or_inst, method_name):
    """
    Returns a method from a given class or instance. When the method doest not exist, it returns `None`. Also works with
    properties and cached properties.
    """
    cls = cls_or_inst if isinstance(cls_or_inst, type) else cls_or_inst.__class__
    meth = getattr(cls, method_name, None)
    if isinstance(meth, property):
        meth = meth.fget
    elif isinstance(meth, cached_property):
        meth = meth.func
    return meth


def keep_spacing(value, autoescape=True):
    """
    When a given `str` contains multiple spaces it keeps first space and others are replaces by &nbsp;. Newlines
    converts into HTML's <br>.
    """
    autoescape = autoescape and not isinstance(value, SafeData)
    value = normalize_newlines(value)
    if autoescape:
        value = escape(value)
    return mark_safe(value.replace('  ', ' &nbsp;').replace('\n', '<br />'))


def call_method_with_unknown_input(method, **fun_kwargs):
    method_kwargs_names = inspect.getargspec(method)[0][1:]

    method_kwargs = {arg_name: fun_kwargs[arg_name] for arg_name in method_kwargs_names if arg_name in fun_kwargs}

    if len(method_kwargs_names) == len(method_kwargs):
        return method(**method_kwargs)
    else:
        raise RuntimeError('Invalid method parameters')



from django import forms
from django.utils.translation import ugettext


class DecimalField(forms.DecimalField):

    def __init__(self, *args, **kwargs):
        self.step = kwargs.pop('step', 'any')
        self.min = kwargs.pop('min', None)
        self.max = kwargs.pop('max', None)
        super(DecimalField, self).__init__(*args, **kwargs)

    def widget_attrs(self, widget):
        attrs = super(DecimalField, self).widget_attrs(widget)
        attrs['step'] = self.step
        if self.min is not None:
            attrs['min'] = self.min
        if self.max is not None:
            attrs['max'] = self.max
        return attrs


class PriceNumberInput(forms.NumberInput):

    def __init__(self, currency, *args, **kwargs):
        super(PriceNumberInput, self).__init__(*args, **kwargs)
        self.placeholder = currency


class PriceField(DecimalField):

    widget = PriceNumberInput

    def __init__(self, *args, **kwargs):
        currency = kwargs.pop('currency', ugettext('CZK'))
        if 'widget' not in kwargs:
            kwargs['widget'] = PriceNumberInput(currency)
        super(PriceField, self).__init__(*args, **kwargs)



from django.core.exceptions import ValidationError
from django.core.files import File
from django.test import TransactionTestCase
from django.utils.translation import ugettext_lazy

from chamber.exceptions import PersistenceException
from chamber.forms import fields as form_fields
from chamber.models.fields import generate_random_upload_path
from chamber.shortcuts import change_and_save

from germanium.anotations import data_provider
from germanium.tools import assert_equal, assert_is_none, assert_false, assert_is_not_none, assert_raises, assert_true

from test_chamber.models import CSVRecord, TestFieldsModel  # pylint: disable=E0401


class ModelFieldsTestCase(TransactionTestCase):

    def setUp(self):
        self.inst = TestFieldsModel()
        super(ModelFieldsTestCase, self).setUp()

    def test_random_file_path_should_be_generated_from_class_name(self):
        instance = CSVRecord()
        filename = 'filename.txt'
        path = generate_random_upload_path(instance, filename)
        assert_true(path.startswith('csvrecord/'))
        assert_true(path.endswith('/{}'.format(filename)))

    def test_decimal_field(self):
        change_and_save(self.inst, decimal=3)
        assert_equal(self.inst.decimal, 3)
        assert_raises(PersistenceException, change_and_save, self.inst, decimal='2.99')
        assert_raises(PersistenceException, change_and_save, self.inst, decimal='10.00001')
        try:
            change_and_save(self.inst, decimal='11.1')
            assert_true(False, 'Previous `change_and_save` suppose to raise an exception')
        except PersistenceException as ex:
            assert_true('decimal: ' in str(ex), 'Exception message was supposed to contain a field name.')

    def test_subchoices_field(self):
        change_and_save(self.inst, state_reason=TestFieldsModel.STATE_REASON.SUB_OK_2)
        assert_equal(self.inst.state_reason, TestFieldsModel.STATE_REASON.SUB_OK_2)
        assert_raises(PersistenceException, change_and_save, self.inst,
                      state_reason=TestFieldsModel.STATE_REASON.SUB_NOT_OK_1)

        # Change state and substate should work
        change_and_save(self.inst, state=TestFieldsModel.STATE.NOT_OK,
                        state_reason=TestFieldsModel.STATE_REASON.SUB_NOT_OK_2)
        assert_equal(self.inst.state, TestFieldsModel.STATE.NOT_OK)
        assert_equal(self.inst.state_reason, TestFieldsModel.STATE_REASON.SUB_NOT_OK_2)

    def test_subchoices_field_value_should_be_empty(self):
        self.inst.state = 4  # setting an invalid value
        try:
            TestFieldsModel._meta.get_field('state_reason').validate(  # pylint: disable=W0212
                TestFieldsModel.STATE_REASON.SUB_NOT_OK_2, self.inst)  # pylint: disable=W0212
            assert_true(False, 'Field validation should raise an error')
        except ValidationError as ex:
            assert_equal(['Value must be empty'], ex.messages)
        assert_is_none(TestFieldsModel._meta.get_field('state_reason').clean(  # pylint: disable=W0212
            TestFieldsModel.STATE_REASON.SUB_NOT_OK_2, self.inst))  # pylint: disable=W0212

    def test_prev_value_field(self):
        # In case of `add`, value of state is copied to state_prev
        change_and_save(self.inst, state=TestFieldsModel.STATE.OK,
                        state_reason=TestFieldsModel.STATE_REASON.SUB_OK_1)
        assert_equal(self.inst.state, TestFieldsModel.STATE.OK)
        assert_equal(self.inst.state_prev, TestFieldsModel.STATE.OK)

        # Later, old value of state is stored in state_prev
        change_and_save(self.inst, state=TestFieldsModel.STATE.NOT_OK,
                        state_reason=TestFieldsModel.STATE_REASON.SUB_NOT_OK_1)
        assert_equal(self.inst.state_prev, TestFieldsModel.STATE.OK)

    def test_sequence_choices_num_enum(self):
        # Only the first state is valid when field is null because it is the initial state
        assert_is_none(self.inst.state_graph)
        assert_raises(PersistenceException, change_and_save, self.inst, state_graph=TestFieldsModel.GRAPH.SECOND)
        assert_raises(PersistenceException, change_and_save, self.inst, state_graph=TestFieldsModel.GRAPH.THIRD)
        change_and_save(self.inst, state_graph=TestFieldsModel.GRAPH.FIRST)
        assert_equal(self.inst.state_graph, TestFieldsModel.GRAPH.FIRST)

        # Than one can switch only to second state
        assert_raises(PersistenceException, change_and_save, self.inst, state_graph=TestFieldsModel.GRAPH.THIRD)
        change_and_save(self.inst, state_graph=TestFieldsModel.GRAPH.SECOND)
        assert_equal(self.inst.state_graph, TestFieldsModel.GRAPH.SECOND)

        # We cannot go back to first but we can go to the third state
        assert_raises(PersistenceException, change_and_save, self.inst, state_graph=TestFieldsModel.GRAPH.FIRST)
        change_and_save(self.inst, state_graph=TestFieldsModel.GRAPH.THIRD)
        assert_equal(self.inst.state_graph, TestFieldsModel.GRAPH.THIRD)

    def test_file_field_content_type(self):
        # These files can be saved because it has supported type
        for filename in ('all_fields_filled.csv', 'test.pdf'):
            with open('data/{}'.format(filename), 'rb') as f:
                assert_false(self.inst.file)
                self.inst.file.save(filename, File(f))
                assert_true(self.inst.file)
                change_and_save(self.inst, file=None)

        # Image file is not supported
        with open('data/test2.jpg', 'rb') as f:
            assert_false(self.inst.file)
            assert_raises(PersistenceException, self.inst.file.save, 'image.jpeg', File(f))

    def test_image_field_max_upload_size(self):
        # File is can be stored
        with open('data/test2.jpg', 'rb') as f:
            assert_false(self.inst.image)
            self.inst.image.save('test2.jpg', File(f))
            assert_true(self.inst.image)
            change_and_save(self.inst, image=None)

        # File is too large to store
        with open('data/test.jpg', 'rb') as f:
            assert_false(self.inst.image)
            assert_raises(PersistenceException, self.inst.image.save, 'image.jpeg', File(f))

        # File has a wrong extension
        with open('data/test2.jpg', 'rb') as f:
            assert_raises(PersistenceException, self.inst.image.save, 'image.html', File(f))

    def test_should_validate_positive_price_field(self):
        assert_raises(PersistenceException, change_and_save, self.inst, total_price=-100)

    def test_should_check_price_form_field(self):
        field = TestFieldsModel._meta.get_field('price')  # pylint: disable=W0212
        assert_equal(ugettext_lazy('EUR'), field.currency)
        form_field = field.formfield()
        assert_true(isinstance(form_field.widget, form_fields.PriceNumberInput))
        assert_equal(field.currency, form_field.widget.placeholder)

    def test_should_check_total_price_form_field(self):
        field = TestFieldsModel._meta.get_field('total_price')  # pylint: disable=W0212
        assert_equal(ugettext_lazy('CZK'), field.currency)
        form_field = field.formfield()
        assert_true(isinstance(form_field.widget, form_fields.PriceNumberInput))

    model_fields = (
        ('price', ugettext_lazy('EUR'), {'max_digits', 'decimal_places'}),
        ('total_price', ugettext_lazy('CZK'), {'max_digits', 'decimal_places', 'validators'}),
    )

    @data_provider(model_fields)
    def test_should_assert_form_field(self, field_name, currency, kwargs_to_remove):
        field = TestFieldsModel._meta.get_field(field_name)  # pylint: disable=W0212
        assert_equal(currency, field.currency)
        form_field = field.formfield()
        assert_true(isinstance(form_field.widget, form_fields.PriceNumberInput))
        assert_equal(field.currency, form_field.widget.placeholder)



import os

from io import StringIO

from django.core.management import call_command
from django.core.serializers import base, python


class MigrationLoadFixture:

    def __init__(self, migration_file, fixture_dir=None, fixture_filename=None, fixture_type='json'):
        self.migration_file = migration_file
        self.fixture_dir = fixture_dir or os.path.abspath(os.path.join(os.path.dirname(migration_file), 'fixtures'))
        self.fixture_filename = fixture_filename or '%s.%s' % (
            os.path.basename(migration_file).rsplit('.', 1)[0], fixture_type
        )

    def __call__(self, apps, schema_editor):
        def _get_model(model_identifier):
            """
            Helper to look up a model from an "app_label.model_name" string.
            """
            try:
                return apps.get_model(model_identifier)
            except (LookupError, TypeError):
                raise base.DeserializationError("Invalid model identifier: '%s'" % model_identifier)

        get_model_tmp = python._get_model  # pylint: disable=W0212
        python._get_model = _get_model
        file = os.path.join(self.fixture_dir, self.fixture_filename)
        if not os.path.isfile(file):
            raise IOError('File "%s" does not exists' % file)
        call_command('loaddata', file, stdout=StringIO())
        python._get_model = get_model_tmp  # pylint: disable=W0212



from six import StringIO

from django.core.management import call_command
from django.test import TestCase

from germanium.tools import assert_equal  # pylint: disable=E0401

from test_chamber.importers import BulkCSVRecordImporter, CSVRecordImporter  # pylint: disable=E0401
from test_chamber.models import CSVRecord  # pylint: disable=E0401


class ImporterTestCase(TestCase):

    def test_records_should_be_bulk_imported_from_csv(self):
        assert_equal(CSVRecord.objects.count(), 0)
        BulkCSVRecordImporter().import_csv()

        assert_equal(CSVRecord.objects.count(), 7)
        assert_equal(CSVRecord.objects.last().name, 'Geordi LaForge')  # Ensure correct value is stored
        assert_equal(CSVRecord.objects.last().number, 888)  # Ensure clean methods work

    def test_records_should_be_imported_without_optional_fields_should_be_bulk_imported_from_csv(self):
        assert_equal(CSVRecord.objects.count(), 0)
        CSVRecordImporter().import_csv(custom_csv_path='data/required_fields_filled.csv')
        assert_equal(CSVRecord.objects.count(), 5)

    def test_records_should_be_imported_from_csv(self):
        assert_equal(CSVRecord.objects.count(), 0)
        BulkCSVRecordImporter().import_csv()

        assert_equal(CSVRecord.objects.count(), 7)
        assert_equal(CSVRecord.objects.last().name, 'Geordi LaForge')  # Ensure correct value is stored
        assert_equal(CSVRecord.objects.last().number, 888)  # Ensure clean methods work

    def test_records_should_be_imported_without_optional_fields_should_be_imported_from_csv(self):
        assert_equal(CSVRecord.objects.count(), 0)
        CSVRecordImporter().import_csv(custom_csv_path='data/required_fields_filled.csv')
        assert_equal(CSVRecord.objects.count(), 5)

    def test_records_should_be_bulk_imported_from_csv_with_command(self):
        assert_equal(CSVRecord.objects.count(), 0)
        call_command('bulk_csv_import', stdout=StringIO(), stderr=StringIO())
        assert_equal(CSVRecord.objects.count(), 7)
        assert_equal(CSVRecord.objects.last().name, 'Geordi LaForge')  # Ensure correct value is stored
        assert_equal(CSVRecord.objects.last().number, 888)  # Ensure clean methods work

    def test_records_should_be_imported_without_optional_fields_should_be_bulk_imported_from_csv_with_command(self):
        assert_equal(CSVRecord.objects.count(), 0)
        call_command('csv_import', stdout=StringIO(), stderr=StringIO())
        assert_equal(CSVRecord.objects.count(), 7)
        assert_equal(CSVRecord.objects.last().name, 'Geordi LaForge')  # Ensure correct value is stored
        assert_equal(CSVRecord.objects.last().number, 888)  # Ensure clean methods work



from django.test import TestCase
from django.utils.functional import cached_property
from django.utils.safestring import SafeData, mark_safe

from chamber.utils import get_class_method, keep_spacing, remove_accent

from germanium.anotations import data_provider  # pylint: disable=E0401
from germanium.tools import assert_equal, assert_true  # pylint: disable=E0401

from .datastructures import *  # NOQA
from .decorators import *  # NOQA


class TestClass(object):

    def method(self):
        pass

    @property
    def property_method(self):
        pass

    @cached_property
    def cached_property_method(self):
        pass


class UtilsTestCase(TestCase):

    def test_should_remove_accent_from_string(self):
        assert_equal('escrzyaie', remove_accent('ěščřžýáíé'))

    classes_and_method_names = [
        [TestClass.method, TestClass, 'method'],
        [TestClass.method, TestClass(), 'method'],
        [TestClass.property_method.fget, TestClass, 'property_method'],
        [TestClass.property_method.fget, TestClass(), 'property_method'],
        [TestClass.cached_property_method.func, TestClass, 'cached_property_method'],
        [TestClass.cached_property_method.func, TestClass(), 'cached_property_method'],
    ]

    @data_provider(classes_and_method_names)
    def test_should_return_class_method(self, expected_method, cls_or_inst, method_name):
        assert_equal(expected_method, get_class_method(cls_or_inst, method_name))

    values_for_keep_spacing = [
        ['Hello &lt;b&gt; escaped&lt;/b&gt; <br />world', 'Hello <b> escaped</b> \nworld', True],
        ['Hello <b> escaped</b> <br />world', 'Hello <b> escaped</b> \nworld', False],
        ['Hello &lt;b&gt; escaped&lt;/b&gt; <br />world', 'Hello <b> escaped</b> \r\nworld', True],
        ['Hello <b> escaped</b> <br />world', 'Hello <b> escaped</b> \r\nworld', False],
        ['Hello <b> escaped</b> <br />world', mark_safe('Hello <b> escaped</b> \r\nworld'), True]
    ]

    @data_provider(values_for_keep_spacing)
    def test_should_keep_spacing(self, expected, value, autoescape):
        escaped_value = keep_spacing(value, autoescape)
        assert_equal(expected, escaped_value)
        assert_true(isinstance(escaped_value, SafeData))



from datetime import date, datetime

from django.core.exceptions import FieldError, MultipleObjectsReturned
from django.http.response import Http404
from django.test import TestCase
from django.utils import timezone

from chamber.shortcuts import (bulk_change, bulk_change_and_save, bulk_save, change, change_and_save, distinct_field,
                               exclude_by_date, filter_by_date, get_object_or_404, get_object_or_none)

from germanium.tools import assert_equal, assert_is_none, assert_raises  # pylint: disable=E0401

from test_chamber.models import ShortcutsModel


class ShortcutsTestCase(TestCase):

    def test_get_object_or_none(self):
        obj = ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=2)
        ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=3)

        assert_equal(get_object_or_none(ShortcutsModel, name='test1'), obj)
        assert_is_none(get_object_or_none(ShortcutsModel, name='test3'))
        assert_is_none(get_object_or_none(ShortcutsModel, number='test3'))
        assert_is_none(get_object_or_none(ShortcutsModel, datetime='test3'))

        assert_raises(FieldError, get_object_or_none, ShortcutsModel, non_field='test2')
        assert_raises(MultipleObjectsReturned, get_object_or_none, ShortcutsModel, name='test2')

    def test_get_object_or_404(self):
        obj = ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=2)
        ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=3)

        assert_equal(get_object_or_404(ShortcutsModel, name='test1'), obj)
        assert_raises(Http404, get_object_or_404, ShortcutsModel, name='test3')
        assert_raises(Http404, get_object_or_404, ShortcutsModel, number='test3')
        assert_raises(Http404, get_object_or_404, ShortcutsModel, datetime='test3')

        assert_raises(FieldError, get_object_or_none, ShortcutsModel, non_field='test2')
        assert_raises(MultipleObjectsReturned, get_object_or_none, ShortcutsModel, name='test2')

    def test_distinct_fields(self):
        ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=2)
        ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=3)
        assert_equal(tuple(distinct_field(ShortcutsModel, 'name')), (('test1',), ('test2',)))
        assert_equal(list(distinct_field(ShortcutsModel, 'name', flat=True)), ['test1', 'test2'])

    def test_filter_and_exclude_by_date(self):
        ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        ShortcutsModel.objects.create(name='test1', datetime=timezone.make_aware(datetime(2014, 1, 1, 12, 12),
                                                                                 timezone.get_default_timezone()),
                                      number=1)
        assert_equal(filter_by_date(ShortcutsModel, datetime=date.today()).count(), 2)
        assert_equal(filter_by_date(ShortcutsModel, datetime=date(2014, 1, 1)).count(), 1)

        assert_equal(exclude_by_date(ShortcutsModel, datetime=date.today()).count(), 1)
        assert_equal(exclude_by_date(ShortcutsModel, datetime=date(2014, 1, 1)).count(), 2)

    def test_change(self):
        obj = ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        change(obj, name='modified')
        assert_equal(obj.name, 'modified')
        assert_equal(ShortcutsModel.objects.first().name, 'test1')  # instance is changed but NOT saved to DB

    def test_change_and_save(self):
        obj = ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        change_and_save(obj, name='modified')
        assert_equal(obj.name, 'modified')
        assert_equal(ShortcutsModel.objects.first().name, 'modified')  # instance is changed and saved to DB

    def test_bulk_change_and_save(self):
        obj1 = ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        obj2 = ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=2)
        bulk_change_and_save([obj1, obj2], name='modified')
        assert_equal(obj1.name, 'modified')
        assert_equal(obj2.name, 'modified')
        assert_equal(ShortcutsModel.objects.first().name, 'modified')  # instance is changed but NOT saved to DB
        assert_equal(ShortcutsModel.objects.last().name, 'modified')  # instance is changed but NOT saved to DB

    def test_bulk_change_and_bulk_save(self):
        obj1 = ShortcutsModel.objects.create(name='test1', datetime=timezone.now(), number=1)
        obj2 = ShortcutsModel.objects.create(name='test2', datetime=timezone.now(), number=2)

        # Bulk change objects
        bulk_change([obj1, obj2], name='modified')
        assert_equal(obj1.name, 'modified')
        assert_equal(obj2.name, 'modified')
        assert_equal(ShortcutsModel.objects.first().name, 'test1')  # instance is changed but NOT saved to DB
        assert_equal(ShortcutsModel.objects.last().name, 'test2')  # instance is changed but NOT saved to DB

        # Bulk save objects
        bulk_save([obj1, obj2])
        assert_equal(ShortcutsModel.objects.first().name, 'modified')  # instance is changed but saved to DB
        assert_equal(ShortcutsModel.objects.last().name, 'modified')  # instance is changed but saved to DB



from django.core.management.base import BaseCommand

from chamber.importers import BulkCSVImporter, CSVImporter

import pyprind


class ProgressBarStream:
    """
    OutputStream wrapper to remove default linebreak at line endings.
    """

    def __init__(self, stream):
        """
        Wrap the given stream.
        """
        self.stream = stream

    def write(self, *args, **kwargs):
        """
        Call the stream's write method without linebreaks at line endings.
        """
        return self.stream.write(ending="", *args, **kwargs)

    def flush(self):
        """
        Call the stream's flush method without any extra arguments.
        """
        return self.stream.flush()


class ImportCSVCommandMixin:

    def handle(self, *args, **kwargs):
        self.import_csv()


class BulkImportCSVCommand(ImportCSVCommandMixin, BulkCSVImporter, BaseCommand):

    def __init__(self, *args, **kwargs):
        super(BulkImportCSVCommand, self).__init__(*args, **kwargs)
        self.bar = None

    def _pre_import_rows(self, row_count):
        self.bar = pyprind.ProgBar(row_count, stream=ProgressBarStream(self.stdout))

    def _post_batch_create(self, created_count, row_count):
        self.bar.update(iterations=created_count)

    def _post_import_rows(self, created_count, updated_count=0):
        self.stdout.write('\nCreated {created} {model_name}.'.format(
            created=created_count,
            model_name=self.model_class._meta.verbose_name_plural  # pylint: disable=W0212
        ))


class ImportCSVCommand(ImportCSVCommandMixin, CSVImporter, BaseCommand):

    def _post_import_rows(self, created_count, updated_count=0):
        self.stdout.write('Created {created} {model_name} and {updated} updated.'.format(
            created=created_count,
            model_name=self.model_class._meta.verbose_name_plural,  # pylint: disable=W0212
            updated=updated_count)
        )



from django.forms import TextInput
from django.test import TransactionTestCase

from germanium.tools import assert_equal, assert_true  # pylint: disable=E0401

from chamber.forms.fields import DecimalField


class DecimalFieldTestCase(TransactionTestCase):

    def test_should_return_correct_widget_attrs(self):
        kwargs = {
            'step': 0.5,
            'min': 1.0,
            'max': 10.0,
        }
        field = DecimalField(**kwargs)
        widget_attrs = field.widget_attrs(TextInput())
        assert_true(len(widget_attrs.keys()) > 0)
        for attr, value in kwargs.items():
            assert_equal(value, widget_attrs[attr])

    def test_should_return_default_attrs(self):
        field = DecimalField()
        widget_attrs = field.widget_attrs(TextInput())
        assert_equal({'step': 'any'}, widget_attrs)



import mimetypes
import os
from decimal import Decimal
from uuid import uuid4 as uuid

import magic  # pylint: disable=E0401

from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import FileField as OriginFileField
from django.db.models.fields import DecimalField as OriginDecimalField
from django.forms import forms
from django.template.defaultfilters import filesizeformat
from django.utils.encoding import force_text
from django.utils.translation import ugettext

from chamber import config
from chamber.forms import fields as chamber_fields
from chamber.models.humanized_helpers import price_humanized
from chamber.utils.datastructures import SequenceChoicesEnumMixin, SubstatesChoicesNumEnum


try:
    from sorl.thumbnail import ImageField as OriginImageField
except ImportError:
    from django.db.models import ImageField as OriginImageField


class DecimalField(OriginDecimalField):

    def __init__(self, *args, **kwargs):
        self.step = kwargs.pop('step', 'any')
        self.min = kwargs.pop('min', None)
        self.max = kwargs.pop('max', None)
        kwargs['validators'] = kwargs.get('validators', [])
        if self.min is not None:
            kwargs['validators'].append(MinValueValidator(self.min))
        if self.max is not None:
            kwargs['validators'].append(MaxValueValidator(self.max))
        super(DecimalField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': chamber_fields.DecimalField,
            'step': self.step,
            'min': self.min,
            'max': self.max,
        }
        defaults.update(kwargs)
        return super(DecimalField, self).formfield(**defaults)


class RestrictedFileValidator:

    def __init__(self, max_upload_size):
        self.max_upload_size = max_upload_size

    def __call__(self, data):
        if data.size > self.max_upload_size:
            raise forms.ValidationError(
                ugettext('Please keep filesize under {max}. Current filesize {current}').format(
                    max=filesizeformat(self.max_upload_size),
                    current=filesizeformat(data.size)
                )
            )
        else:
            return data


class AllowedContentTypesByFilenameFileValidator:

    def __init__(self, content_types):
        self.content_types = content_types

    def __call__(self, data):
        extension_mime_type = mimetypes.guess_type(data.name)[0]

        if extension_mime_type not in self.content_types:
            raise ValidationError(ugettext('Extension of file name is not allowed'))

        return data


class AllowedContentTypesByContentFileValidator:

    def __init__(self, content_types):
        self.content_types = content_types

    def __call__(self, data):
        with magic.Magic(flags=magic.MAGIC_MIME_TYPE) as m:
            mime_type = m.id_buffer(data.file.read(1024))
            data.file.seek(0)
            if mime_type not in self.content_types:
                raise ValidationError(ugettext('File content was evaluated as not supported file type'))

        return data


class RestrictedFileFieldMixin:
    """
    Same as FileField, but you can specify:
        * allowed_content_types - list of allowed content types. Example: ['application/json', 'image/jpeg']
        * max_upload_size - a number indicating the maximum file size allowed for upload in MB.
    """
    def __init__(self, *args, **kwargs):
        max_upload_size = kwargs.pop('max_upload_size', config.CHAMBER_MAX_FILE_UPLOAD_SIZE) * 1024 * 1024
        allowed_content_types = kwargs.pop('allowed_content_types', None)
        super(RestrictedFileFieldMixin, self).__init__(*args, **kwargs)
        self.validators.append(RestrictedFileValidator(max_upload_size))
        if allowed_content_types:
            self.validators = tuple(self.validators) + (
                AllowedContentTypesByFilenameFileValidator(allowed_content_types),
                AllowedContentTypesByContentFileValidator(allowed_content_types),
            )

    def generate_filename(self, instance, filename):
        """
        removes UTF chars from filename
        """
        from unidecode import unidecode

        return super(RestrictedFileFieldMixin, self).generate_filename(instance, unidecode(force_text(filename)))


class FileField(RestrictedFileFieldMixin, OriginFileField):
    pass


class ImageField(RestrictedFileFieldMixin, OriginImageField):

    def __init__(self, *args, **kwargs):
        allowed_content_types = kwargs.pop('allowed_content_types', config.CHAMBER_DEFAULT_IMAGE_ALLOWED_CONTENT_TYPES)
        super(ImageField, self).__init__(allowed_content_types=allowed_content_types, *args, **kwargs)


def generate_random_upload_path(instance, filename):
    """
    Pass this function to upload_to argument of FileField to store the file on an unguessable path.
    The format of the path is class_name/hash/original_filename.
    """
    return os.path.join(instance.__class__.__name__.lower(), uuid().hex, filename)


class PrevValuePositiveIntegerField(models.PositiveIntegerField):

    def __init__(self, *args, **kwargs):
        self.copy_field_name = kwargs.pop('copy_field_name', None)
        super(PrevValuePositiveIntegerField, self).__init__(*args, **kwargs)

    def pre_save(self, model_instance, add):
        if add or hasattr(model_instance, 'changed_fields') and self.copy_field_name in model_instance.changed_fields:
            setattr(
                model_instance, self.attname,
                getattr(model_instance, self.copy_field_name)
                if add else model_instance.initial_values[self.copy_field_name]
            )
        return super(PrevValuePositiveIntegerField, self).pre_save(model_instance, add)


class SubchoicesPositiveIntegerField(models.PositiveIntegerField):

    empty_values = ()

    def __init__(self, *args, **kwargs):
        self.enum = kwargs.pop('enum', None)
        self.subchoices_field_name = kwargs.pop('subchoices_field_name', None)
        assert self.enum is None or isinstance(self.enum, SubstatesChoicesNumEnum)
        if self.enum:
            kwargs['choices'] = self.enum.choices
        super(SubchoicesPositiveIntegerField, self).__init__(*args, **kwargs)

    def _get_subvalue(self, model_instance):
        return getattr(model_instance, self.subchoices_field_name)

    def clean(self, value, model_instance):
        if self.enum and self._get_subvalue(model_instance) not in self.enum.categories:
            return None
        else:
            return super(SubchoicesPositiveIntegerField, self).clean(value, model_instance)

    def _raise_error_if_value_should_be_empty(self, value, subvalue):
        if self.enum and subvalue not in self.enum.categories and value is not None:
            raise ValidationError(ugettext('Value must be empty'))

    def _raise_error_if_value_not_allowed(self, value, subvalue, model_instance):
        allowed_values = self.enum.get_allowed_states(getattr(model_instance, self.subchoices_field_name))
        if subvalue in self.enum.categories and value not in allowed_values:
            raise ValidationError(ugettext('Allowed choices are {}.').format(
                ', '.join(('{} ({})'.format(*(self.enum.get_label(val), val)) for val in allowed_values))
            ))

    def validate(self, value, model_instance):
        if not self.enum:
            return

        self._raise_error_if_value_should_be_empty(value, self._get_subvalue(model_instance))
        self._raise_error_if_value_not_allowed(value, self._get_subvalue(model_instance), model_instance)


class EnumSequenceFieldMixin:

    # TODO Once SmartWidget mixin is not in is-core, add formfield method with the appropriate widget
    def __init__(self, *args, **kwargs):
        self.enum = kwargs.pop('enum', None)
        self.prev_field_name = kwargs.pop('prev_field', None)
        assert self.enum is None or isinstance(self.enum, SequenceChoicesEnumMixin)
        if self.enum:
            kwargs['choices'] = self.enum.choices
        super(EnumSequenceFieldMixin, self).__init__(*args, **kwargs)

    def validate(self, value, model_instance):
        super(EnumSequenceFieldMixin, self).validate(value, model_instance)
        if self.enum:
            prev_value = model_instance.pk and model_instance.initial_values[self.attname] or None
            allowed_next_values = self.enum.get_allowed_next_states(prev_value, model_instance)
            if self.name in model_instance.changed_fields and value not in allowed_next_values:
                raise ValidationError(
                    ugettext('Allowed choices are {}.').format(
                        ', '.join(('{} ({})'.format(*(self.enum.get_label(val), val)) for val in allowed_next_values))))


class EnumSequencePositiveIntegerField(EnumSequenceFieldMixin, models.PositiveIntegerField):
    pass


class EnumSequenceCharField(EnumSequenceFieldMixin, models.CharField):
    pass


class PriceField(DecimalField):

    def __init__(self, *args, **kwargs):
        self.currency = kwargs.pop('currency', ugettext('CZK'))
        default_kwargs = {
            'decimal_places': 2,
            'max_digits': 10,
            'humanized': lambda val, inst, field: price_humanized(val, inst, currency=field.currency)
        }
        default_kwargs.update(kwargs)
        super(PriceField, self).__init__(*args, **default_kwargs)

    def formfield(self, **kwargs):
        default_kwargs = {
            'form_class': chamber_fields.PriceField,
            'currency': self.currency,
        }
        default_kwargs.update(kwargs)

        return super(PriceField, self).formfield(**default_kwargs)


class PositivePriceField(PriceField):

    def __init__(self, *args, **kwargs):
        kwargs['validators'] = kwargs.get('validators', [])
        kwargs['validators'].append(MinValueValidator(Decimal('0.00')))
        super(PositivePriceField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(PositivePriceField, self).deconstruct()
        del kwargs['validators']
        return name, path, args, kwargs



from django.core.management.commands import makemessages


class Command(makemessages.Command):
    extra_args = [
        '--keyword=_l',
        '--keyword=_n:1,2',
        '--keyword=_nl:1,2',
        '--keyword=_p:1c,2',
        '--keyword=_np:1c,2,3',
        '--keyword=_pl:1c,2',
        '--keyword=_npl:1c,2,3',
    ]

    xgettext_options = makemessages.Command.xgettext_options + extra_args



import csv
import io
import os

from django.conf import settings

import pyprind


try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


def simple_count(filename, encoding):
    # Encoding must be passed to match encoding used in the importer.
    lines = 0
    with io.open(filename, encoding=encoding) as f:
        for _ in f:
            lines += 1
    return lines


class DummyOutputStream(io.StringIO):

    def write(self, *args, **kwargs):
        super(DummyOutputStream, self).write(*args)


class AbstractCSVImporter:
    """
    Abstract CSV importer provides an easy way to implement loading a CSV file into a Django model.
        * To alter or validate field value, define clean_field_name() method in a similar manner as in Django forms.
        * The class implements __call__ method to allow calling concrete importers as regular functions.
        * __call__ accepts custom CSV path to import different CSV files with the same instance of the importer.
        * all class properties can be set dynamically with getters.
    """

    skip_header = True  # By default first line of CSV is assumed to contain headers and is skipped
    fields = ()  # Must correspond to columns in the CSV but columns set dynamically in clean methods can be appended.
    csv_path = ''  # Path to the CSV file relative to Django PROJECT_DIR
    delimiter = ';'
    encoding = 'utf-8'

    def __call__(self, csv_path):
        """csv_path is a required parameter as calling the function without CSV does not make sense"""
        self.import_csv(custom_csv_path=csv_path)

    def import_csv(self, custom_csv_path=None):
        with io.open(self.get_filename(custom_csv_path=custom_csv_path), encoding=self.get_encoding()) as f:
            reader = csv.reader(f, delimiter=self.get_delimiter())
            if self.get_skip_header():
                next(reader, None)
            self.import_rows(
                reader,
                row_count=simple_count(self.get_filename(custom_csv_path=custom_csv_path), encoding=self.get_encoding())
            )

    def import_rows(self, reader, row_count=0):
        raise NotImplementedError

    @property
    def out_stream(self):
        """
        By default, output stream is essentially turned off by supplying dummy StringIO.
        Override this property if you want to direct output somewhere, e.g. in Django commands.
        """
        return DummyOutputStream()

    def get_filename(self, custom_csv_path=None):
        return os.path.join(settings.PROJECT_DIR, custom_csv_path or self.get_csv_path())

    def get_csv_path(self):
        """Override this in case you need to set the CSV path dynamically."""
        return self.csv_path

    def get_encoding(self):
        return self.encoding

    def get_delimiter(self):
        return str(self.delimiter)

    def get_skip_header(self):
        return self.skip_header

    def get_fields(self):
        return self.fields

    def get_fields_dict(self, row):
        """
        Returns a dict of field name and cleaned value pairs to initialize the model.
        Beware, it aligns the lists of fields and row values with Nones to allow for adding fields not found in the CSV.
        Whitespace around the value of the cell is stripped.
        """
        return {k: getattr(self, 'clean_{}'.format(k), lambda x: x)(v.strip() if isinstance(v, str)
                                                                    else None)
                for k, v in zip_longest(self.get_fields(), row)}

    def _pre_import_rows(self, row_count):
        pass

    def _post_import_rows(self, created_count, updated_count=0):
        pass


class BulkCSVImporter(AbstractCSVImporter):

    delete_existing_objects = False  # Set to true if you want to delete all records in the model table
    batch_size = 10000

    def create_batch(self, chunk):
        return len(self.model_class.objects.bulk_create(chunk))

    def import_rows(self, reader, row_count=0):
        self._pre_import_rows(row_count)

        if self.get_delete_existing_objects():
            self.model_class.objects.all().delete()

        if self.get_skip_header():
            row_count -= 1

        batch, created = [], 0

        for i, row in enumerate(reader):
            if i % self.get_batch_size() == 0 and i > 0:
                created += self.create_batch(batch)
                self._post_batch_create(self.get_batch_size(), row_count)
                del batch[:]
            if any(row):  # Skip blank lines
                batch.append(self.model_class(**self.get_fields_dict(row)))
        created += self.create_batch(batch)
        self._post_batch_create(len(batch), row_count)

        self._post_import_rows(created)

        return created

    def get_delete_existing_objects(self):
        return self.delete_existing_objects

    def get_batch_size(self):
        return self.batch_size

    def _post_batch_create(self, created_count, row_count):
        pass


class CSVImporter(AbstractCSVImporter):

    query_fields = ()  # Fields used to get existing instances of the model in update_or_create, by default all fields
    update_fields = ()  # Fields that should be set by the update_or_create method

    def import_rows(self, reader, row_count=0):
        self._pre_import_rows(row_count)
        created_flags = [self.row_to_model(row) for row in reader if any(row)]
        self._post_import_rows(sum(created_flags), len(created_flags) - sum(created_flags))

    def row_to_model(self, row):
        fields_dict = self.get_fields_dict(row)
        _, created = self.model_class.objects.update_or_create(
            defaults=self.get_update_dict(fields_dict),
            **self.get_query_dict(fields_dict)
        )
        return created

    def get_query_fields(self):
        return self.query_fields or self.fields

    def get_update_fields(self):
        return self.update_fields

    def get_query_dict(self, fields_dict):
        return {k: v for k, v in fields_dict.items() if k in self.get_query_fields()}

    def get_update_dict(self, fields_dict):
        return {k: v for k, v in fields_dict.items() if k in self.get_update_fields()}



import re

from collections import MutableSet, OrderedDict
from itertools import chain


ENUM_KEY_PATTERN = re.compile(r'^[a-zA-Z][a-zA-Z0-9_]*$')


class AbstractEnum:

    def __init__(self, *items):
        for k, _ in items:
            if not isinstance(k, str):
                raise ValueError('Enum key "{}" must be string'.format(k))
            if not ENUM_KEY_PATTERN.match(k):
                raise ValueError('Enum key "{}" has invalid format'.format(k))

        self._container = OrderedDict(items)
        self._reverse_container = {item[1]: item[0] for item in items}

    def _has_attr(self, name):
        return name in self._container

    def __getattr__(self, name):
        if self._has_attr(name):
            return self._container[name]
        raise AttributeError('Missing attribute %s' % name)

    def __copy__(self, *args, **kwargs):
        # Enum is immutable
        return self

    def __deepcopy__(self, *args, **kwargs):
        # Enum is immutable
        return self

    def __contains__(self, item):
        return item in self._reverse_container

    def __iter__(self):
        return self._container.values().__iter__()

    @property
    def all(self):
        return tuple(self)

    def get_name(self, val):
        return self._reverse_container.get(val)


class Enum(AbstractEnum):

    def __init__(self, *items):
        super().__init__(*(
            item if isinstance(item, (list, tuple)) else (item, item)
            for item in items
        ))


class NumEnum(Enum):

    def __init__(self, *items):
        used_ids = set()

        enum_items = []
        i = 0
        for item in items:
            if len(item) == 2:
                key, i = item
                if not isinstance(i, int):
                    raise ValueError('Choice value of item must by integer')
            else:
                key = item
                i += 1

            if i in used_ids:
                raise ValueError('Index %s already exists, please renumber choices')

            used_ids.add(i)
            enum_items.append((key, i))
        super().__init__(*enum_items)


class AbstractChoicesEnum:

    def __init__(self, *items):
        enum_items = []
        for item in items:
            assert len(item) in {2, 3}, 'Choice item array length must be two or three'

            if len(item) == 3:
                enum_items.append((item[0], item[2]))
            else:
                enum_items.append(item[0])

        super().__init__(*enum_items)
        self.choices = tuple(
            (k, items[i][1]) for i, k in enumerate(self._container.values())
        )

    def _get_labels_dict(self):
        return dict(self.choices)

    def get_label(self, name):
        labels = self._get_labels_dict()
        if name in labels:
            return labels[name]
        raise AttributeError('Missing label with index %s' % name)


class ChoicesEnum(AbstractChoicesEnum, Enum):
    pass


class ChoicesNumEnum(AbstractChoicesEnum, NumEnum):
    pass


class SubstatesChoicesNumEnum(ChoicesNumEnum):

    def __init__(self, categories):
        assert len(categories) > 0

        self.categories = {}

        super(SubstatesChoicesNumEnum, self).__init__(*(item for item in chain(*categories.values())))

        self.categories = {
            category: [getattr(self, item[0]) for item in subitems] for category, subitems in categories.items()
        }

    def get_allowed_states(self, category):
        return self.categories.get(category, ())


class SequenceChoicesEnumMixin:

    def __init__(self, items, initial_states=None):
        assert len(items) > 0

        self.initial_states = initial_states

        # The last value of every item are omitted and send to ChoicesEnum constructor
        super().__init__(*(item[:-1] for item in items if item[0] is not None))

        self.first_choices = self._get_first_choices(items)

        # The last value of every item is used for construction of graph that define allowed next states for every state
        self.sequence_graph = {getattr(self, item[0]): item[-1] for item in items}

    def _get_first_choices(self, items):
        return tuple(getattr(self, key) for key in self.initial_states) if self.initial_states else self.all

    def get_allowed_next_states(self, state, instance):
        if not state:
            return self.first_choices
        else:
            states_or_callable = self.sequence_graph.get(state)
            states = (states_or_callable(instance) if hasattr(states_or_callable, '__call__')
                      else list(states_or_callable))
            return tuple(getattr(self, next_choice) for next_choice in states)


class SequenceChoicesEnum(SequenceChoicesEnumMixin, ChoicesEnum):
    pass


class SequenceChoicesNumEnum(SequenceChoicesEnumMixin, ChoicesNumEnum):
    pass


class OrderedSet(MutableSet):

    def __init__(self, *iterable):
        self.end = end = []
        end += [None, end, end]  # sentinel node for doubly linked list
        self.map = {}  # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev_item, next_item = self.map.pop(key)
            prev_item[2] = next_item
            next_item[1] = prev_item

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '{}()'.format(self.__class__.__name__,)
        else:
            return '{}({})'.format(self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)


class AttrString(str):

    def __new__(cls, value, **kwargs):
        obj = str.__new__(cls, value)
        [setattr(obj, k, v) for k, v in kwargs.items()]
        return obj



from chamber.importers import BulkCSVImporter, CSVImporter

from .models import CSVRecord


class BulkCSVRecordImporter(BulkCSVImporter):
    model_class = CSVRecord
    fields = ('id', 'name', 'number')
    csv_path = 'data/all_fields_filled.csv'

    def clean_number(self, value):
        # Just to test clean methods are called
        return 888


class CSVRecordImporter(CSVImporter):
    model_class = CSVRecord
    fields = ('id', 'name', 'number')
    csv_path = 'data/all_fields_filled.csv'

    def clean_number(self, value):
        # Just to test clean methods are called
        return 888



from setuptools import setup, find_packages

from chamber.version import get_version


setup(
    name="django-chamber",
    version=get_version(),
    description="Utilities library meant as a complement to django-is-core.",
    author="Lubos Matl, Oskar Hollmann",
    author_email="matllubos@gmail.com, oskar@hollmann.me",
    url="http://github.com/druids/django-chamber",
    packages=find_packages(),
    package_dir={"chamber": "chamber"},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Framework :: Django",
    ],
    install_requires=[
        'Django>=1.8',
        'Unidecode>=0.04.17',
        'pyprind>=2.11.0',
        'six>=1.10.0',
        'filemagic>=1.6',
    ],
)



import logging

from collections import OrderedDict

from django.conf import settings
from django.db import transaction, DEFAULT_DB_ALIAS
from django.db.transaction import get_connection
from django.utils.decorators import ContextDecorator


logger = logging.getLogger(__name__)


def atomic(func):
    """
    Decorator helper that overrides django atomic decorator and automatically adds create revision.
    """
    try:
        from reversion.revisions import create_revision

        return transaction.atomic(create_revision()(func))
    except ImportError:
        return transaction.atomic(func)


class TransactionSignalsContext:
    """
    Context object that stores handlers and call it after successful pass trough surrounded code block
    with "transaction_signals decorator. Handlers can be unique or standard. Unique handlers are registered
    and executed only once.
    """

    def __init__(self):
        self._unique_handlers = OrderedDict()
        self._handlers = []

    def register(self, handler):
        if getattr(handler, 'is_unique', False):
            if hash(handler) in self._unique_handlers:
                self._unique_handlers.get(hash(handler)).join(handler)
            else:
                self._unique_handlers[hash(handler)] = handler
                self._handlers.append(handler)
        else:
            self._handlers.append(handler)

    def handle_all(self):
        for handler in self._handlers:
            handler()

    def join(self, transaction_signals_context):
        for handler in transaction_signals_context._handlers:
            self.register(handler)


class TransactionSignals(ContextDecorator):
    """
    Context decorator that supports usage python keyword "with".
    Decorator that adds transaction context to the connection on input.
    Finally handlers are called on the output.
    """

    def __init__(self, using):
        self.using = using

    def __enter__(self):
        connection = get_connection(self.using)

        if not hasattr(connection, 'transaction_signals_context_list'):
            connection.transaction_signals_context_list = []

        connection.transaction_signals_context_list.append(TransactionSignalsContext())

    def __exit__(self, exc_type, exc_value, traceback):
        connection = get_connection(self.using)
        transaction_signals_context = connection.transaction_signals_context_list.pop()
        if not exc_value:
            if len(connection.transaction_signals_context_list) == 0:
                transaction_signals_context.handle_all()
            else:
                connection.transaction_signals_context_list[-1].join(transaction_signals_context)


def on_success(handler, using=None):
    """
    Register a handler or a function to be called after successful code pass.
    If transaction signals are not active the handler/function is called immediately.
    :param handler: handler or function that will be called.
    :param using: name of the database
    """

    connection = get_connection(using)
    if getattr(connection, 'transaction_signals_context_list', False):
        connection.transaction_signals_context_list[-1].register(handler)
    else:
        if settings.DEBUG:
            logger.warning(
                'For on success signal should be activated transaction signals via transaction_signals decorator.'
                'Function is called immediately now.'
            )
        handler()


def transaction_signals(using=None):
    """
    Decorator that adds transaction context to the connection on input.
    Finally handlers are called on the output.
    :param using: name of the database
    """
    if callable(using):
        return TransactionSignals(DEFAULT_DB_ALIAS)(using)
    else:
        return TransactionSignals(using)


def atomic_with_signals(func):
    """
    Atomic decorator with transaction signals.
    """
    try:
        from reversion.revisions import create_revision

        return transaction.atomic(create_revision()(transaction_signals(func)))
    except ImportError:
        return transaction.atomic(transaction_signals(func))


class OnSuccessHandler:
    """
    Handler class that is used for performing on success operations.
    """

    is_unique = False

    def __init__(self, using=None, **kwargs):
        self.kwargs = kwargs
        on_success(self, using=using)

    def __call__(self):
        self.handle(**self.kwargs)

    def handle(self, **kwargs):
        """
        There should be implemented handler operations.
        :param kwargs: input data that was send during hanlder creation.
        """
        raise NotImplementedError


class OneTimeOnSuccessHandler(OnSuccessHandler):
    """
    One time handler class that is used for performing on success operations.
    Handler is called only once, but data of all calls are stored inside list (kwargs_list).
    """

    is_unique = True

    def __init__(self, using=None, **kwargs):
        self.kwargs_list = (kwargs,)
        on_success(self, using=using)

    def join(self, handler):
        """
        Joins two unique handlers.
        """
        self.kwargs_list += handler.kwargs_list

    def _get_unique_id(self):
        """
        Unique handler must be identified with some was
        :return:
        """
        return None

    def __hash__(self):
        return hash((self.__class__, self._get_unique_id()))

    def __call__(self):
        self.handle(self.kwargs_list)

    def handle(self, kwargs_list):
        raise NotImplementedError


class InstanceOneTimeOnSuccessHandler(OneTimeOnSuccessHandler):
    """
    Use this class to create handler that will be unique per instance and will be called only once per instance.
    """

    def _get_instance(self):
        instance = self.kwargs_list[0]['instance']
        instance.refresh_from_db()
        return instance

    def _get_unique_id(self):
        instance = self.kwargs_list[0]['instance']
        return hash((instance.__class__, instance.pk))



class PersistenceException(Exception):

    def __init__(self, message=None):
        super(PersistenceException, self).__init__()
        self.message = message

    def __str__(self):
        return repr(self.message)



import copy

import collections

from itertools import chain

from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.db.models.base import ModelBase
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import force_text

from chamber.exceptions import PersistenceException
from chamber.patch import Options
from chamber.shortcuts import change_and_save, change, bulk_change_and_save

from .fields import *  # NOQA exposing classes and functions as a module API
from .signals import dispatcher_post_save, dispatcher_pre_save


def many_to_many_field_to_dict(field, instance):
    if instance.pk is None:
        # If the object doesn't have a primary key yet, just use an empty
        # list for its m2m fields. Calling f.value_from_object will raise
        # an exception.
        return []
    else:
        # MultipleChoiceWidget needs a list of pks, not object instances.
        return list(field.value_from_object(instance).values_list('pk', flat=True))


def should_exclude_field(field, fields, exclude):
    return (fields and field.name not in fields) or (exclude and field.name in exclude)


def field_to_dict(field, instance):
    """
    Converts a model field to a dictionary
    """
    # avoid a circular import
    from django.db.models.fields.related import ManyToManyField

    return (many_to_many_field_to_dict(field, instance) if isinstance(field, ManyToManyField)
            else field.value_from_object(instance))


def model_to_dict(instance, fields=None, exclude=None):
    """
    The same implementation as django model_to_dict but editable fields are allowed
    """

    return {
        field.name: field_to_dict(field, instance)
        for field in chain(instance._meta.concrete_fields, instance._meta.many_to_many)  # pylint: disable=W0212
        if not should_exclude_field(field, fields, exclude)
    }


ValueChange = collections.namedtuple('ValueChange', ('initial', 'current'))


class ChangedFields:
    """
    Class stores changed fields and its initial and current values.
    """

    def __init__(self, initial_dict):
        self._initial_dict = initial_dict

    @property
    def initial_values(self):
        return self._initial_dict

    @property
    def current_values(self):
        raise NotImplementedError

    @property
    def diff(self):
        d1 = self.initial_values
        d2 = self.current_values
        return {k: ValueChange(v, d2[k]) for k, v in d1.items() if v != d2[k]}

    def __setitem__(self, key, item):
        raise AttributeError('Object is readonly')

    def __getitem__(self, key):
        return self.diff[key]

    def __repr__(self):
        return repr(self.diff)

    def __len__(self):
        return len(self.diff)

    def __delitem__(self, key):
        raise AttributeError('Object is readonly')

    def clear(self):
        raise AttributeError('Object is readonly')

    def has_key(self, k):
        return k in self.diff

    def has_any_key(self, *keys):
        return bool(set(self.keys()) & set(keys))

    def update(self, *args, **kwargs):
        raise AttributeError('Object is readonly')

    def keys(self):
        return self.diff.keys()

    def values(self):
        return self.diff.values()

    def items(self):
        return self.diff.items()

    def pop(self, *args, **kwargs):
        raise AttributeError('Object is readonly')

    def __cmp__(self, dictionary):
        return cmp(self.diff, dictionary)

    def __contains__(self, item):
        return item in self.diff

    def __iter__(self):
        return iter(self.diff)

    def __str__(self):
        return repr(self.diff)


class DynamicChangedFields(ChangedFields):
    """
    Dynamic changed fields are changed with the instance changes.
    """

    def __init__(self, instance):
        super(DynamicChangedFields, self).__init__(self._get_instance_dict(instance))
        self.instance = instance

    def _get_instance_dict(self, instance):
        return model_to_dict(instance, fields=(field.name for field in instance._meta.fields))

    @property
    def current_values(self):
        return self._get_instance_dict(self.instance)

    def get_static_changes(self):
        return StaticChangedFields(self.initial_values, self.current_values)


class StaticChangedFields(ChangedFields):
    """
    Static changed fields are immutable. The origin instance changes will not have an affect.
    """

    def __init__(self, initial_dict, current_dict):
        super(StaticChangedFields, self).__init__(initial_dict)
        self._current_dict = current_dict

    @property
    def current_values(self):
        return self._current_dict


class ComparableModelMixin:

    def equals(self, obj, comparator):
        """
        Use comparator for evaluating if objects are the same
        """
        return comparator.compare(self, obj)


class Comparator:

    def compare(self, a, b):
        """
        Return True if objects are same otherwise False
        """
        raise NotImplementedError


class AuditModel(models.Model):

    created_at = models.DateTimeField(verbose_name=_('created at'), null=False, blank=False, auto_now_add=True,
                                      db_index=True)
    changed_at = models.DateTimeField(verbose_name=_('changed at'), null=False, blank=False, auto_now=True,
                                      db_index=True)

    class Meta:
        abstract = True


class Signal:

    def __init__(self, obj):
        self.connected_functions = []
        self.obj = obj

    def connect(self, fun):
        self.connected_functions.append(fun)

    def send(self):
        [fun(self.obj) for fun in self.connected_functions]


class SmartQuerySet(models.QuerySet):

    def fast_distinct(self):
        """
        Because standard distinct used on the all fields are very slow and works only with PostgreSQL database
        this method provides alternative to the standard distinct method.
        :return: qs with unique objects
        """
        return self.model.objects.filter(pk__in=self.values_list('pk', flat=True))

    def change_and_save(self, **chaned_fields):
        """
        Changes a given `changed_fields` on each object in the queryset, saves objects
        and returns the changed objects in the queryset.
        """
        bulk_change_and_save(self, **chaned_fields)
        return self.filter()


class SmartModelBase(ModelBase):
    """
    Smart model meta class that register dispatchers to the post or pre save signals.
    """

    def __new__(cls, name, bases, attrs):

        new_cls = super(SmartModelBase, cls).__new__(cls, name, bases, attrs)
        for dispatcher in new_cls.dispatchers:
            dispatcher.connect(new_cls)
        return new_cls


class SmartModel(AuditModel, metaclass=SmartModelBase):

    objects = SmartQuerySet.as_manager()

    dispatchers = []

    def __init__(self, *args, **kwargs):
        super(SmartModel, self).__init__(*args, **kwargs)
        self.changed_fields = DynamicChangedFields(self)
        self.post_save = Signal(self)

    @property
    def has_changed(self):
        return bool(self.changed_fields)

    @property
    def initial_values(self):
        return self.changed_fields.initial_values

    def full_clean(self, exclude=None, *args, **kwargs):
        errors = {}
        for field in self._meta.fields:
            if (not exclude or field.name not in exclude) and hasattr(self, 'clean_{}'.format(field.name)):
                try:
                    getattr(self, 'clean_{}'.format(field.name))()
                except ValidationError as er:
                    errors[field.name] = er

        if errors:
            raise ValidationError(errors)
        super(SmartModel, self).full_clean(exclude=exclude, *args, **kwargs)

    def _clean_save(self, *args, **kwargs):
        self._persistence_clean(*args, **kwargs)

    def _clean_delete(self, *args, **kwargs):
        self._persistence_clean(*args, **kwargs)

    def _clean_pre_save(self, *args, **kwargs):
        self._clean_save(*args, **kwargs)

    def _clean_pre_delete(self, *args, **kwargs):
        self._clean_delete(*args, **kwargs)

    def _clean_post_save(self, *args, **kwargs):
        self._clean_save(*args, **kwargs)

    def _clean_post_delete(self, *args, **kwargs):
        self._clean_delete(*args, **kwargs)

    def _persistence_clean(self, *args, **kwargs):
        exclude = kwargs.pop('exclude', None)
        try:
            self.full_clean(exclude=exclude)
        except ValidationError as er:
            if hasattr(er, 'error_dict'):
                raise PersistenceException(', '.join(
                    ('%s: %s' % (key, ', '.join(map(force_text, val))) for key, val in er.message_dict.items())))
            else:
                raise PersistenceException(', '.join(map(force_text, er.messages)))

    def _get_save_extra_kwargs(self):
        return {}

    def _pre_save(self, *args, **kwargs):
        pass

    def _call_pre_save(self, *args, **kwargs):
        self._pre_save(*args, **kwargs)

    def _save(self, is_cleaned_pre_save=None, is_cleaned_post_save=None, force_insert=False, force_update=False,
              using=None, update_fields=None, *args, **kwargs):

        is_cleaned_pre_save = (
            self._smart_meta.is_cleaned_pre_save if is_cleaned_pre_save is None else is_cleaned_pre_save
        )
        is_cleaned_post_save = (
            self._smart_meta.is_cleaned_post_save if is_cleaned_post_save is None else is_cleaned_post_save
        )

        origin = self.__class__

        change = bool(self.pk)
        kwargs.update(self._get_save_extra_kwargs())

        self._call_pre_save(change, self.changed_fields, *args, **kwargs)
        if is_cleaned_pre_save:
            self._clean_pre_save(*args, **kwargs)
        dispatcher_pre_save.send(sender=origin, instance=self, change=change,
                                 changed_fields=self.changed_fields.get_static_changes(),
                                 *args, **kwargs)
        super(SmartModel, self).save(force_insert=force_insert, force_update=force_update, using=using,
                                     update_fields=update_fields)
        self._call_post_save(change, self.changed_fields, *args, **kwargs)
        if is_cleaned_post_save:
            self._clean_post_save(*args, **kwargs)
        dispatcher_post_save.send(sender=origin, instance=self, change=change,
                                  changed_fields=self.changed_fields.get_static_changes(),
                                  *args, **kwargs)
        self.post_save.send()

    def _post_save(self, *args, **kwargs):
        pass

    def _call_post_save(self, *args, **kwargs):
        self._post_save(*args, **kwargs)

    def save(self, *args, **kwargs):
        if self._smart_meta.is_save_atomic:
            with transaction.atomic():
                self._save(*args, **kwargs)
        else:
            self._save(*args, **kwargs)
        self.changed_fields = DynamicChangedFields(self)

    def _pre_delete(self, *args, **kwargs):
        pass

    def _delete(self, is_cleaned_pre_delete=None, is_cleaned_post_delete=None, *args, **kwargs):
        is_cleaned_pre_delete = (
            self._smart_meta.is_cleaned_pre_delete if is_cleaned_pre_delete is None else is_cleaned_pre_delete
        )
        is_cleaned_post_delete = (
            self._smart_meta.is_cleaned_post_delete if is_cleaned_post_delete is None else is_cleaned_post_delete
        )

        self._pre_delete(*args, **kwargs)

        if is_cleaned_pre_delete:
            self._clean_pre_delete(*args, **kwargs)

        super(SmartModel, self).delete(*args, **kwargs)

        self._post_delete(*args, **kwargs)

        if is_cleaned_post_delete:
            self._clean_post_delete(*args, **kwargs)

    def _post_delete(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        if self._smart_meta.is_delete_atomic:
            with transaction.atomic():
                self._delete(*args, **kwargs)
        else:
            self._delete(*args, **kwargs)

    def refresh_from_db(self, *args, **kwargs):
        super().refresh_from_db(*args, **kwargs)
        return self

    def change(self, **changed_fields):
        """
        Changes a given `changed_fields` on this object and returns itself.
        :param changed_fields: fields to change
        :return: self
        """
        change(self, **changed_fields)
        return self

    def change_and_save(self, **changed_fields):
        """
        Changes a given `changed_fields` on this object, saves it and returns itself.
        :param changed_fields: fields to change
        :return: self
        """
        change_and_save(self, **changed_fields)
        return self

    class Meta:
        abstract = True


class SmartOptions(Options):

    meta_class_name = 'SmartMeta'
    meta_name = '_smart_meta'
    model_class = SmartModel
    attributes = {
        'is_cleaned_pre_save': True,
        'is_cleaned_post_save': False,
        'is_cleaned_pre_delete': False,
        'is_cleaned_post_delete': False,
        'is_save_atomic': False,
        'is_delete_atomic': False,
    }
